function [dict, W, output]=BaselineTrain(dir_name, file_type, param)
%cvpr 2008
%Author
%   Qiang Zhang
%   CIDSE, Arizona State University
%   qzhang53@asu.edu
%Description:
%   Just as the method in [1], but instead of manually choosing the dictionary, we use the
%   images to train the dictionary. First, we have Ni images for L people. Then we run the KSVD
%   for Ni images of every single people i, and obtain a dictionary of size Mi for i. All
%   these Mi sized dictionary consists the global dictionary D. Then for the test image, we
%   sparse code it, with method such as OMP, and get the coefficient A. We can identify the
%   test image buy find the i that minimizes ||y-deta(i, D*A)||.
%Parameter:
%   dir_name:  the name of directory where you store the training set;
%   file_type: type of file you want to use as training set;
%   param:
%              no_people: number of the people in the training set;
%              no_image:  number of images in the training set for every people;
%              no_atom:   number of atoms in the trained dictionary for all people;
%              support:   the non zero elements in the coeffcient;
%              sample:    the random sample matrix, or resize [y x];
%              lable:     the label with the trainging image;
%              iteration: number of interations;
%              alpha:     weight for class and dict, suggest alpha=255;
%   dict:      output the trained dictionary. The atom is organized as columns every atom is a trained picture.
%   W:         classifier based on coefficient of training images.
%   output:    contains the coefficient matrix;
%Recommendation:
% param = 
% 
%     no_people: 38
%      no_image: 19
%       no_atom: 304
%       support: 8
%        sample: [504x32256 double]
%         lable: []
%     iteration: 50
%Caution: 
%   Please set your directory first, because this function willprocess all the image in the directory.
%   All the images should be 8 bit gray and take the same size

%go to the directory
cd(dir_name);
% get a dir list of all images of type 'imagetype'
eval( sprintf('dirlist=dir(''*.%s'');', file_type ) ); 

%build up the training images
if size(param.sample, 1)>1
    %has to do some random sample
    img=imread(dirlist(1).name);
    len_img=size(img, 1)*size(img, 2);
    data=zeros(size(param.sample, 1), length(dirlist));
    for i=1:length(dirlist)
        img=imread(dirlist(i).name);
        if size(img,3)>1
            img=rgb2gray(img);
        end
        data(:, i)=param.sample*double(reshape(img, len_img, 1));
    end
else
    %first get the size of images
    len_img=param.sample(1)*param.sample(2);
    data=zeros(param.sample(1)*param.sample(2), length(dirlist));
    for i=1:length(dirlist)
        img=imread(dirlist(i).name);
        if size(img,3)>1
            img=rgb2gray(img);
        end
        img=imresize(img, param.sample);
        data(:, i)=double(reshape(img, len_img, 1));
    end
end

%set up parameters for ksvd training
param_train.L=param.support;
param_train.K=param.no_atom;
param_train.numIteration=20;
param_train.errorFlag=0;
param_train.errorGoal=4;
param_train.preserveDCAtom=0;
param_train.InitializationMethod =  'DataElements';
param_train.displayProgress = 1;
%=========================================================================
%stage 1
%train the images together with ksvd
%=========================================================================
[dict, ini_output]=KSVD(data, param_train);
%some other parameters
I=diag(ones(size(dict, 2)));
I=diag(I);
if isempty(param.lable)==0
    Y=param.lable;
else
    Y=zeros(param.no_people, length(dirlist));
    for i=1:param.no_people
        Y(i, (i-1)*param.no_image+1:i*param.no_image)=1;
    end
end
totalErr = zeros(1,param.iteration);
% the weight for classify and represent
LDratio=param.alpha;
Data=[LDratio*Y; data];
coef=ini_output.CoefMatrix;
% W=((coef*coef'+I)\coef*Y')';
%=========================================================================
%iterations
for i=1:param.iteration
    %=========================================================================
    %stage 2
    %use the dict and output to find the classfier
    %=========================================================================
    %use w=(Z*Z'+gammar*I)^(-1)*Z*Y;
    W=((coef*coef'+I)\coef*Y')';
    %=========================================================================
    %stage 3
    %update the dictionary
    %find the solution to (D Z)=argmin||[Y; LDratio*X]-Z[W; LDratio*D]||2
    %while keep the sparsity of Z. It is very similar to KSVD
    %=========================================================================
    replacedVectorCounter = 0;
    Dictionary=[LDratio*W; dict];
    coef=OMP(Dictionary, Data, param.support);
    rPerm = randperm(size(Dictionary,2));
    %begin to update the D and Z, while fixing W
    for j = rPerm
        [betterDictionaryElement,coef,addedNewVector] = I_findBetterDictionaryElement(Data, Dictionary,j, coef ,param.support);
        Dictionary(:,j) = betterDictionaryElement;
    end
    %output, W is unchanged
    Dictionary = I_clearDictionary(Dictionary,coef,Data);
    dict=Dictionary(size(W,1)+1: size(Dictionary,1), :);
    err= sqrt(sum(sum((data-dict*coef).^2)))/prod(size(data));
    disp(err);
    %display
    disp(['Iteration: ', num2str(i)]);
    %=========================================================================
end
output.CoefMatrix = coef;
%=========================================================================
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  findBetterDictionaryElement
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%something different from the paper
function [betterDictionaryElement,CoefMatrix,NewVectorAdded] = I_findBetterDictionaryElement(Data,Dictionary,j,CoefMatrix,numCoefUsed)
%j is the index of atom at dictionary
if (length(who('numCoefUsed'))==0)
    numCoefUsed = 1;
end
relevantDataIndices = find(CoefMatrix(j,:)); % the data indices that uses the j'th dictionary element.
%find(x) will return the linear index of nonzero element in x
if (length(relevantDataIndices)<1) %(length(relevantDataIndices)==0)
    %none of the signal use atom j
    ErrorMat = Data-Dictionary*CoefMatrix;
    %Ek=Y-D*X
    ErrorNormVec = sum(ErrorMat.^2);
    [d,i] = max(ErrorNormVec);
    betterDictionaryElement = Data(:,i);%ErrorMat(:,i); %
    betterDictionaryElement = betterDictionaryElement./sqrt(betterDictionaryElement'*betterDictionaryElement);
    betterDictionaryElement = betterDictionaryElement.*sign(betterDictionaryElement(1));
    %ERk=Ek*Wk
    %Wk(i,j)=(wk(j),j), j is the index of the atom
    CoefMatrix(j,:) = 0;
    NewVectorAdded = 1;
    return;
end

%this is the svd section
NewVectorAdded = 0;
%Wk
tmpCoefMatrix = CoefMatrix(:,relevantDataIndices); 
tmpCoefMatrix(j,:) = 0;% the coeffitients of the element we now improve are not relevant.
%ERk
errors =(Data(:,relevantDataIndices) - Dictionary*tmpCoefMatrix); % vector of errors that we want to minimize with the new element
% % the better dictionary element and the values of beta are found using svd.
% % This is because we would like to minimize || errors - beta*element ||_F^2. 
% % that is, to approximate the matrix 'errors' with a one-rank matrix. This
% % is done using the largest singular value.
[betterDictionaryElement,singularValue,betaVector] = svds(errors,1);
CoefMatrix(j,relevantDataIndices) = singularValue*betaVector';% *signOfFirstElem

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  findDistanseBetweenDictionaries
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ratio,totalDistances] = I_findDistanseBetweenDictionaries(original,new)
% first, all the column in oiginal starts with positive values.
catchCounter = 0;
totalDistances = 0;
for i = 1:size(new,2)
    new(:,i) = sign(new(1,i))*new(:,i);
end
for i = 1:size(original,2)
    d = sign(original(1,i))*original(:,i);
    distances =sum ( (new-repmat(d,1,size(new,2))).^2);
    [minValue,index] = min(distances);
    errorOfElement = 1-abs(new(:,index)'*d);
    totalDistances = totalDistances+errorOfElement;
    catchCounter = catchCounter+(errorOfElement<0.01);
end
ratio = 100*catchCounter/size(original,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  I_clearDictionary
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Dictionary = I_clearDictionary(Dictionary,CoefMatrix,Data)
T2 = 0.99;
T1 = 3;
K=size(Dictionary,2);
Er=sum((Data-Dictionary*CoefMatrix).^2,1); % remove identical atoms
G=Dictionary'*Dictionary; G = G-diag(diag(G));
for jj=1:1:K,
    if max(G(jj,:))>T2 | length(find(abs(CoefMatrix(jj,:))>1e-7))<=T1 ,
        [val,pos]=max(Er);
        Er(pos(1))=0;
        Dictionary(:,jj)=Data(:,pos(1))/norm(Data(:,pos(1)));
        G=Dictionary'*Dictionary; G = G-diag(diag(G));
    end;
end;


